﻿//---------------------------------------------------------------------------- 
//
//  Copyright (C) CSharp Labs.  All rights reserved.
// 
//  Permission is hereby granted, free of charge, to any person obtaining a copy
//  of this software and associated documentation files (the "Software"), to deal
//  in the Software without restriction, including without limitation the rights
//  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//  copies of the Software, and to permit persons to whom the Software is
//  furnished to do so, subject to the following conditions:
// 
//  The above copyright notice and this permission notice shall be included in
//  all copies or substantial portions of the Software.
// 
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//  THE SOFTWARE.
// 
// History
//  05/30/13    Created 
//
//---------------------------------------------------------------------------

namespace System.Collections
{
    /// <summary>
    /// The ActionTypeSwitch class allows a type to invoke methods.
    /// </summary>
    public class ActionTypeSwitch : ObjectSwitch<Type, Action<Type>>
    {
        /// <summary>
        /// Invokes a method by looking up the associated case from the specified type.
        /// </summary>
        /// <param name="type">The type to search for.</param>
        public void Switch(Type type)
        {
            //gets associated delegate or default
            Action<Type> del = LookupDelegate(type);

            if (del != null)
                del(type);
        }

        /// <summary>
        /// Boxes the delegate.
        /// </summary>
        /// <param name="del">The delegate to box.</param>
        /// <returns>The boxed delegate.</returns>
        protected override Delegate BoxDelegate(Action<Type> del)
        {
            return del;
        }

        /// <summary>
        /// Unboxes the delegate.
        /// </summary>
        /// <param name="del">The delegate to unbox.</param>
        /// <returns>The unboxed delegate.</returns>
        protected override Action<Type> UnboxDelegate(Delegate del)
        {
            return (Action<Type>)del;
        }
    }

    /// <summary>
    /// The ActionTypeSwitch class allows a type to invoke methods.
    /// </summary>
    public class ActionTypeSwitch<T1> : ObjectSwitch<Type, Action<Type, T1>>
    {
        /// <summary>
        /// Invokes a method by looking up the associated case from the specified type.
        /// </summary>
        /// <param name="type">The type to search for.</param>
        public void Switch(Type type, T1 arg1)
        {
            Action<Type, T1> del = LookupDelegate(type);

            if (del != null)
                del(type, arg1);
        }

        /// <summary>
        /// Boxes the delegate.
        /// </summary>
        /// <param name="del">The delegate to box.</param>
        /// <returns>The boxed delegate.</returns>
        protected override Delegate BoxDelegate(Action<Type, T1> del)
        {
            return del;
        }

        /// <summary>
        /// Unboxes the delegate.
        /// </summary>
        /// <param name="del">The delegate to unbox.</param>
        /// <returns>The unboxed delegate.</returns>
        protected override Action<Type, T1> UnboxDelegate(Delegate del)
        {
            return (Action<Type, T1>)del;
        }
    }

    /// <summary>
    /// The ActionTypeSwitch class allows a type to invoke methods.
    /// </summary>
    public class ActionTypeSwitch<T1, T2> : ObjectSwitch<Type, Action<Type, T1, T2>>
    {
        /// <summary>
        /// Invokes a method by looking up the associated case from the specified type.
        /// </summary>
        /// <param name="type">The type to search for.</param>
        public void Switch(Type type, T1 arg1, T2 arg2)
        {
            Action<Type, T1, T2> del = LookupDelegate(type);

            if (del != null)
                del(type, arg1, arg2);
        }

        /// <summary>
        /// Boxes the delegate.
        /// </summary>
        /// <param name="del">The delegate to box.</param>
        /// <returns>The boxed delegate.</returns>
        protected override Delegate BoxDelegate(Action<Type, T1, T2> del)
        {
            return del;
        }

        /// <summary>
        /// Unboxes the delegate.
        /// </summary>
        /// <param name="del">The delegate to unbox.</param>
        /// <returns>The unboxed delegate.</returns>
        protected override Action<Type, T1, T2> UnboxDelegate(Delegate del)
        {
            return (Action<Type, T1, T2>)del;
        }
    }

    /// <summary>
    /// The ActionTypeSwitch class allows a type to invoke methods.
    /// </summary>
    public class ActionTypeSwitch<T1, T2, T3> : ObjectSwitch<Type, Action<Type, T1, T2, T3>>
    {
        /// <summary>
        /// Invokes a method by looking up the associated case from the specified type.
        /// </summary>
        /// <param name="type">The type to search for.</param>
        public void Switch(Type type, T1 arg1, T2 arg2, T3 arg3)
        {
            Action<Type, T1, T2, T3> del = LookupDelegate(type);

            if (del != null)
                del(type, arg1, arg2, arg3);
        }

        /// <summary>
        /// Boxes the delegate.
        /// </summary>
        /// <param name="del">The delegate to box.</param>
        /// <returns>The boxed delegate.</returns>
        protected override Delegate BoxDelegate(Action<Type, T1, T2, T3> del)
        {
            return del;
        }

        /// <summary>
        /// Unboxes the delegate.
        /// </summary>
        /// <param name="del">The delegate to unbox.</param>
        /// <returns>The unboxed delegate.</returns>
        protected override Action<Type, T1, T2, T3> UnboxDelegate(Delegate del)
        {
            return (Action<Type, T1, T2, T3>)del;
        }
    }
}
