﻿//---------------------------------------------------------------------------- 
//
//  Copyright (C) CSharp Labs.  All rights reserved.
// 
//  Permission is hereby granted, free of charge, to any person obtaining a copy
//  of this software and associated documentation files (the "Software"), to deal
//  in the Software without restriction, including without limitation the rights
//  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//  copies of the Software, and to permit persons to whom the Software is
//  furnished to do so, subject to the following conditions:
// 
//  The above copyright notice and this permission notice shall be included in
//  all copies or substantial portions of the Software.
// 
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//  THE SOFTWARE.
// 
// History
//  05/30/13    Created 
//
//---------------------------------------------------------------------------

namespace System.Collections
{
    using System.Collections.Generic;
    using System.Runtime.Serialization;

    /// <summary>
    /// The ObjectSwitch class allows an object to invoke methods.
    /// </summary>
    /// <typeparam name="TSwitch">The type of object to compare.</typeparam>
    /// <typeparam name="TDelegate">The type of delegate to box or unbox.</typeparam>
    public abstract class ObjectSwitch<TSwitch, TDelegate> where TDelegate : class, ICloneable, ISerializable
    {
        /// <summary>
        /// The static initializer provides a way to restrict TDelegate to delegate types.
        /// </summary>
        static ObjectSwitch()
        {
            if (!typeof(TDelegate).IsSubclassOf(typeof(Delegate)))
                throw new InvalidOperationException(typeof(TDelegate).Name + " is not a delegate type.");
        }

        /// <summary>
        /// Defines a dictionary of cases and actions.
        /// </summary>
        private Dictionary<TSwitch, TDelegate> _TypeRefs;

        /// <summary>
        /// Initializes the TypeSwitch class.
        /// </summary>
        protected ObjectSwitch()
        {
            _TypeRefs = new Dictionary<TSwitch, TDelegate>();
        }

        /// <summary>
        /// Adds a key and associates it with the specified delegate.
        /// </summary>
        /// <param name="newKey">The key to associate with a delegate.</param>
        /// <param name="value">The delegate to invoke.</param>
        /// <exception cref="ArgumentNullException">newKey is null.</exception>
        /// <exception cref="ArgumentNullException">value is null.</exception>
        public void Add(TSwitch newKey, TDelegate value)
        {
            if (newKey == null)
                throw new ArgumentNullException("newKey");
            if (value == null)
                throw new ArgumentNullException("value");

            TDelegate current_value;
            if (_TypeRefs.TryGetValue(newKey, out current_value)) //if the key exists
                _TypeRefs[newKey] = UnboxDelegate(Delegate.Combine(BoxDelegate(current_value), BoxDelegate(value))); //combine delegates
            else //otherwise, add the key and delegate
                _TypeRefs[newKey] = value;
        }

        /// <summary>
        /// Removes a key and delegate.
        /// </summary>
        /// <param name="remKey">The key to remove.</param>
        /// <param name="value">The delegate to remove.</param>
        /// <exception cref="ArgumentNullException">remKey is null.</exception>
        /// <exception cref="ArgumentNullException">value is null.</exception>
        public void Remove(TSwitch remKey, TDelegate value)
        {
            if (remKey == null)
                throw new ArgumentNullException("remKey");

            if (value == null)
                throw new ArgumentNullException("value");

            TDelegate current_value;
            if (_TypeRefs.TryGetValue(remKey, out current_value)) //try to get the delegate
            {
                Delegate boxed_value = BoxDelegate(value); //boxes value to a delegate
                Delegate boxed_current_value = BoxDelegate(current_value); //boxes current value to a delegate
                Delegate[] invocations = boxed_current_value.GetInvocationList(); //get current invocation list

                if (invocations.Length - 1 == 0) //if there is only 1 delegate
                {
                    if (invocations[0] == boxed_current_value) //check if value
                        _TypeRefs.Remove(remKey); //drop from references
                }
                else //otherwise, remove the delegate
                    _TypeRefs[remKey] = UnboxDelegate(Delegate.Remove(boxed_current_value, boxed_value));
            }
        }

        /// <summary>
        /// Determines if the TypeSwitch contains the specified key.
        /// </summary>
        /// <param name="key">The value.</param>
        /// <returns>true if the TypeSwitch contains the specified value; otherwise, false.</returns>
        /// <exception cref="ArgumentNullException">key is null.</exception>
        public bool Contains(TSwitch key)
        {
            if (key == null)
                throw new ArgumentNullException("key");

            return _TypeRefs.ContainsKey(key);
        }

        /// <summary>
        /// Gets or sets the default function to invoke.
        /// </summary>
        public TDelegate Default { get; set; }

        /// <summary>
        /// Looks up the specified key and returns an associated delegate.
        /// </summary>
        /// <param name="key">The key to look up the corresponding delegate from.</param>
        /// <returns>If the key exists, returns the corresponding delegate; if <see cref="Default"/> is not null, returns <see cref="Default"/>; otherwise, returns null.</returns>
        /// <exception cref="ArgumentNullException">key is null.</exception>
        protected TDelegate LookupDelegate(TSwitch key)
        {
            if (key == null)
                throw new ArgumentNullException("key");

            TDelegate value;
            if (_TypeRefs.TryGetValue(key, out value)) //try to get the value
                return value; //return the current value

            //otherwise return the default
            return Default;
        }

        /// <summary>
        /// Boxes the delegate.
        /// </summary>
        /// <param name="del">The delegate to box.</param>
        /// <returns>The boxed delegate.</returns>
        protected abstract Delegate BoxDelegate(TDelegate del);

        /// <summary>
        /// Unboxes the delegate.
        /// </summary>
        /// <param name="del">The delegate to unbox.</param>
        /// <returns>The unboxed delegate.</returns>
        protected abstract TDelegate UnboxDelegate(Delegate del);
    }
}
