﻿//---------------------------------------------------------------------------- 
//
//  Copyright (C) CSharp Labs.  All rights reserved.
// 
//  Permission is hereby granted, free of charge, to any person obtaining a copy
//  of this software and associated documentation files (the "Software"), to deal
//  in the Software without restriction, including without limitation the rights
//  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//  copies of the Software, and to permit persons to whom the Software is
//  furnished to do so, subject to the following conditions:
// 
//  The above copyright notice and this permission notice shall be included in
//  all copies or substantial portions of the Software.
// 
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//  THE SOFTWARE.
// 
// History
//  07/15/13    Created 
//
//---------------------------------------------------------------------------

namespace System.Web.Security
{
    using System.Collections;
    using System.Collections.Generic;
    using System.IO;
    using System.IO.Compression;
    using System.Linq;
    using System.Text;

    /// <summary>
    /// The FormsAuthenticationCookie class creates or modifies <see cref="FormsAuthenticationTicket"/>s to add compressed user data.
    /// </summary>
    public sealed class FormsAuthenticationCookie
    {
        #region DataType Enum
        /// <summary>
        /// Defines types of data the FormsAuthenticationCookie can store.
        /// </summary>
        private enum DataType : byte
        {
            Char = 0,
            String = 1,
            Bool = 2,
            Int32 = 3,
            Int64 = 4,
            DateTime = 5,
        }
        #endregion

        #region Static Fields
        /// <summary>
        /// Defines a type switch to set user data.
        /// </summary>
        private static readonly ActionTypeSwitch<string, Dictionary<string, object>, object> _SetTypeSwitch = new ActionTypeSwitch<string, Dictionary<string, object>, object>();
        /// <summary>
        /// Defines a type switch to write data.
        /// </summary>
        private static readonly ActionTypeSwitch<BinaryWriter, object> _WriteObjectSwitch = new ActionTypeSwitch<BinaryWriter, object>();
        #endregion

        #region Static Constructor
        /// <summary>
        /// Initializes static type switches.
        /// </summary>
        static FormsAuthenticationCookie()
        {
            #region Writing Object Data
            //case DataType: 
            _WriteObjectSwitch.Add(typeof(DataType), (type, writer, value) =>
            {
                //write type of data:
                writer.Write((byte)value);
            });

            //case bool:
            _WriteObjectSwitch.Add(typeof(bool), (type, writer, value) =>
            {
                //write DataType:
                _WriteObjectSwitch.Switch(typeof(DataType), writer, DataType.Bool);
                //write Boolean:
                writer.Write((bool)value);
            });

            //case Int32:
            _WriteObjectSwitch.Add(typeof(Int32), (type, writer, value) =>
            {
                //write DataType:
                _WriteObjectSwitch.Switch(typeof(DataType), writer, DataType.Int32);
                //write integer:
                writer.Write((Int32)value);
            });

            //case Int64:
            _WriteObjectSwitch.Add(typeof(Int64), (type, writer, value) =>
            {
                //write DataType:
                _WriteObjectSwitch.Switch(typeof(DataType), writer, DataType.Int64);
                //write long:
                writer.Write((Int64)value);
            });

            //case char:
            _WriteObjectSwitch.Add(typeof(char), (type, writer, value) =>
            {
                //write DataType:
                _WriteObjectSwitch.Switch(typeof(DataType), writer, DataType.Char);
                //write char:
                writer.Write((char)value);
            });

            //case DateTime:
            _WriteObjectSwitch.Add(typeof(DateTime), (type, writer, value) =>
            {
                //write DataType:
                _WriteObjectSwitch.Switch(typeof(DataType), writer, DataType.DateTime);
                //write serialized DateTime:
                writer.Write(((DateTime)value).ToBinary());
            });

            //case string:
            _WriteObjectSwitch.Add(typeof(string), (type, writer, value) =>
            {
                //write DataType:
                _WriteObjectSwitch.Switch(typeof(DataType), writer, DataType.String);
                //write string:
                writer.Write((string)value);
            });

            //default:
            _WriteObjectSwitch.Default = (type, builder, value) =>
            {
                throw new NotSupportedException("The type '" + type.FullName + "' is not supported.");
            };
            #endregion

            #region Set Dictionary Value
            //defines a method that sets a dictionary value
            Action<Type, string, Dictionary<string, object>, object> setValue = (Type type, string key, Dictionary<string, object> userData, object value) =>
            {
                userData[key] = value;
            };

            //add cases for each type of acceptable data:
            _SetTypeSwitch.Add(typeof(int), setValue);
            _SetTypeSwitch.Add(typeof(Int64), setValue);
            _SetTypeSwitch.Add(typeof(bool), setValue);
            _SetTypeSwitch.Add(typeof(char), setValue);
            _SetTypeSwitch.Add(typeof(DateTime), setValue);
            _SetTypeSwitch.Add(typeof(string), setValue);

            //default to handle invalid object types:
            _SetTypeSwitch.Default = (Type type, string key, Dictionary<string, object> userData, object value) =>
            {
                throw new NotSupportedException("The type '" + type.FullName + "' is not supported.");
            };
            #endregion
        }
        #endregion

        #region Static Properties
        /// <summary>
        /// Creates a new instance of the FormsAuthenticationCookie for an authenticated user or null if the user is not authenticated.
        /// </summary>
        public static FormsAuthenticationCookie Current
        {
            get
            {
                //get the current context:
                HttpContext context = HttpContext.Current;
                FormsAuthenticationTicket ticket;

                //check if cookie exists in response without creating it
                if (context.Response.Cookies.AllKeys.Contains(FormsAuthentication.FormsCookieName))
                {
                    //if existing authentication cookie in response, get the cookie and decrypt the FormsAuthenticationTicket
                    HttpCookie existing = context.Response.Cookies[FormsAuthentication.FormsCookieName];
                    ticket = FormsAuthentication.Decrypt(existing.Value);
                }
                //otherwise, if the user is authenticated:
                else if (context.User != null && context.User.Identity.IsAuthenticated && context.User.Identity is FormsIdentity)
                    //gets the FormsAuthenticationTicket from the user identity:
                    ticket = ((FormsIdentity)context.User.Identity).Ticket;
                else
                    //not authenticated
                    return null;

                //return new FormsAuthenticationCookie
                return new FormsAuthenticationCookie(ticket);
            }
        }
        #endregion

        #region Fields
        /// <summary>
        /// The existing FormsAuthenticationTicket.
        /// </summary>
        private FormsAuthenticationTicket _Ticket;
        /// <summary>
        /// The user if this will create a new FormsAuthenticationTicket.
        /// </summary>
        private string _User;
        /// <summary>
        /// A value determining if the FormsAuthenticationTicket will be persistent if this will create a new FormsAuthenticationTicket.
        /// </summary>
        private bool _IsPersistent;
        /// <summary>
        /// User data to encrypt and store in the FormsAuthenticationTicket.
        /// </summary>
        private Dictionary<string, object> _UserData = new Dictionary<string, object>();
        #endregion

        #region Properties
        /// <summary>
        /// Gets or sets if the user data should be compressed.
        /// </summary>
        public bool CompressUserData
        {
            get;
            set;
        }
        #endregion

        #region Get / Set Value
        /// <summary>
        /// Throws an ArgumentException if the specified key is invalid.
        /// </summary>
        /// <param name="key">The key to validate.</param>
        private static void ThrowIfInvalidKey(string key)
        {
            if (key == null)
                throw new ArgumentNullException("key");

            if (key.Length == 0)
                throw new ArgumentException("Must specify a valid key name.", "key");
        }

        /// <summary>
        /// Determines if the specified key has an associated value.
        /// </summary>
        /// <param name="key">The key to check.</param>
        /// <returns>true if the key has an associated value.</returns>
        public bool ContainsKey(string key)
        {
            return _UserData.ContainsKey(key);
        }

        /// <summary>
        /// Gets or sets the element with the specified key.
        /// </summary>
        /// <param name="key">The key of the element to get or set.</param>
        /// <returns>The element with the specified key.</returns>
        public object this[string key]
        {
            get
            {
                object obj;
                if (_UserData.TryGetValue(key, out obj))
                    return obj;
                else
                    return null;
            }
            set
            {
                ThrowIfInvalidKey(key);

                if (value == null)
                    _UserData.Remove(key);
                else
                    _SetTypeSwitch.Switch(value.GetType(), key, _UserData, value);
            }
        }

        /// <summary>
        /// Gets the value with the specified key.
        /// </summary>
        /// <typeparam name="T">The type of element to get.</typeparam>
        /// <param name="key">The key of the element to get.</param>
        /// <returns>The value associated with the specified key.</returns>
        public T Get<T>(string key)
        {
            return (T)this[key];
        }

        /// <summary>
        /// Sets the specified key and value in the cookie.
        /// </summary>
        /// <typeparam name="T">The type of element to set.</typeparam>
        /// <param name="key">The key of the element to set.</param>
        /// <param name="value">The value of the element to set.</param>
        public void Set<T>(string key, T value)
        {
            this[key] = value;
        }
        #endregion

        #region Constructors
        /// <summary>
        /// Initializes the FormsAuthenticationCookie from an existing FormsAuthenticationTicket.
        /// </summary>
        private FormsAuthenticationCookie(FormsAuthenticationTicket ticket)
        {
            _Ticket = ticket;
            _User = ticket.Name;
            _IsPersistent = ticket.IsPersistent;

            ReadFromTicket();
        }

        /// <summary>
        /// Creates a new FormsAuthenticationCookie for a user.
        /// </summary>
        /// <param name="user">The name of an authenticated user.</param>
        /// <param name="isPersistent">true to create a durable cookie (one that is saved across browser sessions); otherwise, false. </param>
        public FormsAuthenticationCookie(string user, bool isPersistent)
        {
            CompressUserData = true;
            _User = user;
            _IsPersistent = isPersistent;
        }

        /// <summary>
        /// Loads the data from the ticket.
        /// </summary>
        private void ReadFromTicket()
        {
            string userdata = _Ticket.UserData;

            if (string.IsNullOrEmpty(userdata))
                CompressUserData = true;
            else
            {
                //convert base=64 encoded data to binary
                byte[] binary = Convert.FromBase64String(userdata);

                //output stream is used to read uncompressed data
                MemoryStream output = null;

                try
                {
                    if (binary[0] == 1) //if compression flag
                    {
                        //using compression
                        CompressUserData = true;

                        //create stream to compressed data
                        using (MemoryStream compressed = new MemoryStream(binary))
                        {
                            //skip initial compression flag byte
                            compressed.Seek(1, SeekOrigin.Current);

                            //create deflate stream to decompress data:
                            using (DeflateStream deflate = new DeflateStream(compressed, CompressionMode.Decompress, true))
                            {
                                //create output stream
                                output = new MemoryStream();

                                //decompress data to output
                                deflate.CopyTo(output);

                                //seek to beginning of stream
                                output.Seek(0, SeekOrigin.Begin);
                            }
                        }
                    }
                    else
                    {
                        //no compression
                        CompressUserData = false;

                        //create output stream
                        output = new MemoryStream(binary);

                        //skip initial compression flag byte
                        output.Seek(1, SeekOrigin.Current);
                    }

                    using (BinaryReader reader = new BinaryReader(output, Encoding.UTF8, true))
                    {
                        do
                        {
                            string key = reader.ReadString();
                            DataType type = (DataType)reader.ReadByte();
                            object value;

                            switch (type)//get the property data type
                            {
                                case DataType.Bool:
                                    //read Boolean:
                                    value = reader.ReadBoolean();
                                    break;
                                case DataType.Int32:
                                    //read integer:
                                    value = reader.ReadInt32();
                                    break;
                                case DataType.Int64:
                                    //read long:
                                    value = reader.ReadInt64();
                                    break;
                                case DataType.String:
                                    //read string:
                                    value = reader.ReadString();
                                    break;
                                case DataType.Char:
                                    //read char
                                    value = reader.ReadChar();
                                    break;
                                case DataType.DateTime:
                                    //read long and convert to date time:
                                    value = DateTime.FromBinary(reader.ReadInt64());
                                    break;
                                default:
                                    throw new NotSupportedException("Type not supported.");
                            }

                            //add data:
                            _UserData.Add(key, value);
                        }
                        while (output.Position < output.Length); //read until end of stream
                    }
                }
                finally
                {
                    if (output != null)
                        output.Dispose();
                }
            }
        }
        #endregion

        #region Methods
        /// <summary>
        /// Saves or updates the FormsAuthenticationTicket in a HttpCookie.
        /// </summary>
        public void Save()
        {
            string encodedUserData;

            //stream to write binary data to:
            using (MemoryStream stream = new MemoryStream())
            {
                //if this stream will not be compressed
                if (!CompressUserData)
                    //write no compression flag
                    stream.WriteByte(0);

                //write all the values in the user data dictionary
                using (BinaryWriter writer = new BinaryWriter(stream, Encoding.UTF8, true))
                {
                    //write each key and value
                    foreach (KeyValuePair<string, object> item in _UserData)
                    {
                        //write the item key
                        writer.Write(item.Key);
                        //write the item value
                        _WriteObjectSwitch.Switch(item.Value.GetType(), writer, item.Value);
                    }
                }

                //if stream should be compressed
                if (CompressUserData)
                {
                    //seek to the beginning
                    stream.Seek(0, SeekOrigin.Begin);

                    //create a stream to output compressed data to
                    using (MemoryStream output = new MemoryStream())
                    {
                        //write the compression flag
                        output.WriteByte(1);

                        //compress stream using deflate algorithm
                        using (DeflateStream deflate = new DeflateStream(output, CompressionLevel.Optimal, true))
                            stream.CopyTo(deflate);

                        //convert binary to base-64 encoded string
                        encodedUserData = Convert.ToBase64String(output.ToArray());
                    }
                }
                else
                    //convert binary to base-64 encoded string
                    encodedUserData = Convert.ToBase64String(stream.ToArray());
            }

            DateTime issueDate, expirationDate;
            int version;
            string cookiePath;

            if (_Ticket == null)
            {
                //ticket issue date
                issueDate = DateTime.UtcNow;
                //ticket expiration date
                expirationDate = issueDate.Add(FormsAuthentication.Timeout);
                //use version 2, default FormsAuthenticationTicket.Version
                version = 2;
                //ticket cookie path
                cookiePath = FormsAuthentication.FormsCookiePath;
            }
            else //otherwise, use original values
            {
                //get ticket issue date
                issueDate = _Ticket.IssueDate;
                //get ticket expiration date
                expirationDate = _Ticket.Expiration;
                //get ticket version
                version = _Ticket.Version;
                //get ticket cookie path
                cookiePath = _Ticket.CookiePath;
            }

            //create a new ticket:
            FormsAuthenticationTicket ticket = new FormsAuthenticationTicket(version, _User, issueDate, expirationDate, _IsPersistent, encodedUserData, cookiePath);

            //create a new cookie
            HttpCookie cookie = new HttpCookie(FormsAuthentication.FormsCookieName);
            //encrypt ticket
            cookie.Value = FormsAuthentication.Encrypt(ticket);
            //set expiration
            cookie.Expires = ticket.Expiration;
            //set ssl requirements:
            cookie.Secure = FormsAuthentication.RequireSSL;
            //no script access:
            cookie.HttpOnly = true;

            //get current http context
            HttpContext context = HttpContext.Current;

            //if no ticket exists and cookie is not set in response
            if (_Ticket == null && !context.Response.Cookies.AllKeys.Contains(FormsAuthentication.FormsCookieName))
                //add the cookie
                context.Response.Cookies.Add(cookie);
            else //otherwise
                //set the cookie
                context.Response.Cookies.Set(cookie); //set cookie

            _Ticket = ticket;
        }
        #endregion
    }
}